@@ -116,6 +116,7 @@ pub struct Settings<'a> {
 
     pub filter_keys: HashSet<Vec<u8>>,
     pub filter_labels: HashSet<Vec<u8>>,
+    pub filter_null_keys: bool,
 }
 
 impl Default for Settings<'_> {
@@ -139,6 +140,7 @@ impl Default for Settings<'_> {
             unlabel_script: None,
             filter_keys: HashSet::new(),
             filter_labels: HashSet::new(),
+            filter_null_keys: false,
         }
     }
 }
@@ -305,6 +307,16 @@ fn translate_socketaddr(rv: &mut Record, sa: SocketAddr) -> Value {
 /// Returns a script name from path if exe's dev / inode don't match
 ///
 /// This seems to work with Docker containers but not with Podman.
+/// The executable's device and inode are inspected throguh the
+/// /proc/<pid>/root/ symlink. This may fail for
+///
+/// - very short-lived processes
+/// - container setups where the container's filesystem is constructed
+///   using fuse-overlayfs (observed with
+///   podman+fuse-overlayfs/1.4.0-1 on Debian/buster).
+///
+/// As an extra sanity check, exe is compared with normalized
+/// PATH.name. If they are equal, no script is returned.
 fn path_script_name(path: &Record, pid: u32, cwd: &[u8], exe: &[u8]) -> Option<NVec> {
     let mut proc_exe_path = Vec::from(format!("/proc/{}/root", pid).as_bytes());
     proc_exe_path.extend(exe);
@@ -360,7 +372,7 @@ fn path_script_name(path: &Record, pid: u32, cwd: &[u8], exe: &[u8]) -> Option<N
     }
     match (p_dev, p_inode, name) {
         (Some(p_dev), Some(p_inode), _) if p_dev == e_dev && p_inode == e_inode => None,
-        (Some(_), Some(_), Some(name)) => Some(name),
+        (Some(_), Some(_), Some(name)) if name != exe => Some(name),
         _ => None,
     }
 }
@@ -702,22 +714,24 @@ impl<'a> Coalesce<'a> {
         // parent if applicable
         let parent: Option<Process> = ppid.and_then(|ppid| self.processes.get_process(ppid));
 
-        if let (Some(pid), Some(ppid), true) = (pid, ppid, syscall_is_exec) {
-            self.processes.add_process(
-                pid,
-                ppid,
-                ev.id,
-                comm.as_ref().map(|s| s.to_vec()),
-                exe.as_ref().map(|s| s.to_vec()),
-            );
+        if let (Some(pid), Some(ppid)) = (pid, ppid) {
+            if syscall_is_exec || self.processes.get_process(pid).is_none() {
+                self.processes.add_process(
+                    pid,
+                    ppid,
+                    ev.id,
+                    comm.as_ref().map(|s| s.to_vec()),
+                    exe.as_ref().map(|s| s.to_vec()),
+                );
 
-            if let Some(parent) = &parent {
-                for l in self
-                    .settings
-                    .proc_propagate_labels
-                    .intersection(&parent.labels)
-                {
-                    self.processes.add_label(pid, l);
+                if let Some(parent) = &parent {
+                    for l in self
+                        .settings
+                        .proc_propagate_labels
+                        .intersection(&parent.labels)
+                    {
+                        self.processes.add_label(pid, l);
+                    }
                 }
             }
         }
@@ -779,6 +793,9 @@ impl<'a> Coalesce<'a> {
                 ev.filter = true;
                 return;
             }
+        } else if self.settings.filter_null_keys {
+            ev.filter = true;
+            return;
         }
 
         let proc: Option<Process> = pid.and_then(|pid| self.processes.get_process(pid));
@@ -1037,6 +1054,11 @@ impl<'a> Coalesce<'a> {
         Ok(())
     }
 
+    /// Flush all in-flight event data, including partial events
+    pub fn flush(&mut self) {
+        self.expire_inflight(u64::MAX);
+    }
+
     pub fn dump_state(&self, mut w: &mut dyn Write) -> Result<(), Box<dyn Error>> {
         serde_json::to_writer(
             &mut w,
@@ -1075,7 +1097,7 @@ impl<'a> Coalesce<'a> {
 
 impl Drop for Coalesce<'_> {
     fn drop(&mut self) {
-        self.expire_inflight(u64::MAX)
+        self.flush();
     }
 }
 
@@ -1127,7 +1149,14 @@ mod test {
     where
         T: AsRef<[u8]>,
     {
-        for line in BufReader::new(text.as_ref()).lines() {
+        for line in BufReader::new(text.as_ref())
+            .lines()
+            .filter(|line| match line {
+                Ok(l) if l.len() == 0 => false,
+                Ok(l) if l.starts_with("#") => false,
+                _ => true,
+            })
+        {
             let mut line = line.unwrap().clone();
             line.push('\n');
             c.process_line(line.as_bytes().to_vec())?;
@@ -1355,6 +1384,15 @@ mod test {
         drop(c);
         assert!(ec.borrow().as_ref().is_none());
 
+        let mut c = Coalesce::new(emitter);
+        c.settings.filter_null_keys = true;
+        process_record(
+            &mut c,
+            include_bytes!("testdata/record-syscall-nullkey.txt"),
+        )?;
+        drop(c);
+        assert!(ec.borrow().as_ref().is_none());
+
         let mut c = Coalesce::new(emitter);
         c.settings
             .filter_keys
@@ -1474,4 +1512,43 @@ mod test {
 
         Ok(())
     }
+
+    #[test]
+    fn shell_proc_trace() {
+        let events: Rc<RefCell<Vec<Event>>> = Rc::new(RefCell::new(vec![]));
+        let mut c = Coalesce::new(|e| events.borrow_mut().push(e.clone()));
+
+        c.settings.proc_label_keys = [b"test-script".to_vec()].into();
+        c.settings.proc_propagate_labels = [b"test-script".to_vec()].into();
+
+        process_record(&mut c, include_bytes!("testdata/shell-proc-trace.txt")).unwrap();
+
+        let events = events.borrow();
+
+        let fork_ev = events
+            .iter()
+            .find(|e| e.id.to_string() == "1682609045.530:29241")
+            .unwrap();
+        assert!(event_to_json(&fork_ev).contains(r#""LABELS":["test-script"]"#));
+        let script_ev = events
+            .iter()
+            .find(|e| e.id.to_string() == "1682609045.526:29238")
+            .unwrap();
+        assert!(event_to_json(&script_ev).contains(r#""LABELS":["test-script"]"#));
+        let grep_ev = events
+            .iter()
+            .find(|e| e.id.to_string() == "1682609045.530:29242")
+            .unwrap();
+        assert!(event_to_json(&grep_ev).contains(r#""LABELS":["test-script"]"#));
+        let echo_ev = events
+            .iter()
+            .find(|e| e.id.to_string() == "1682609045.530:29244")
+            .unwrap();
+        assert!(event_to_json(&echo_ev).contains(r#""LABELS":["test-script"]"#));
+        let sed_ev = events
+            .iter()
+            .find(|e| e.id.to_string() == "1682609045.534:29245")
+            .unwrap();
+        assert!(event_to_json(&sed_ev).contains(r#""LABELS":["test-script"]"#));
+    }
 }
